#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
# NGAS Utility

import hashlib
import logging
import os
import random
import shutil
import time
import uuid
from urllib.parse import quote_plus

import jxmlease
import requests
from pycapo import CapoConfig

from . import evlasciencedatamodel as sc

capoprofile = os.environ["CAPO_PROFILE"]
config = CapoConfig(profile=capoprofile)
# Retrieve NGAS host information
NGASHOSTS = [x.strip() for x in config.getstring("archive-ingestion.NGASHosts").split(" ")]
archivecommand = {"status": "STATUS", "archive": "ARCHIVE", "retrieve": "RETRIEVE"}
LOG = logging.getLogger("ingestion")


def random_ngas_host():
    return random.choice(NGASHOSTS)


def ingestevlasdmsintongas(filesetbasepath, bdfpath):
    filesets = []
    for fileset in filesets():
        filesetpath = os.path.join(filesetbasepath, fileset)
        print("Processing fileset: ".format(fileset))


def ingestsdmintongas(sdm, sdmpath, bdfpath, ngas_staging):
    # Copy SDM table files
    filelist = []
    for sdmtable in sdm.sdmtables:
        table = sdm.sdmtables[sdmtable]
        if isinstance(sdm.sdmtables[sdmtable], sc.SDMXMLTable):
            fromextension = ".xml"
            toextension = ".sdm"
        elif isinstance(sdm.sdmtables[sdmtable], sc.SDMBinaryTable):
            fromextension = ".bin"
            toextension = ".bin"
        else:
            LOG.critical("Unrecognizable SDM table type")
            exit(-1)

        fromfile = os.path.join(sdmpath, table.tablename + fromextension)
        tofilename = table.entityid.replace(":", "_").replace("/", "_") + toextension
        tofile = os.path.join(ngas_staging, tofilename)
        LOG.info(f"SDM ingestion copying from {fromfile} to {tofile}")
        shutil.copy(fromfile, tofile)
        filelist.append((tofile, os.stat(tofile).st_size))

    # Link BDF files
    for bdffile in sdm.bdffiles:
        extension = ".bdf"
        filename = bdffile.entityid.replace(":", "_").replace("/", "_")
        # fromfile = os.path.join(sdmpath, filename) + extension
        fromfile = os.path.join(bdfpath, filename)
        tofile = os.path.join(ngas_staging, filename + extension)
        # TODO don't copy bdfs yet - disk space
        LOG.info(f"BDF ingestion linking from {fromfile} to {tofile}")
        os.link(fromfile, tofile)
        filelist.append((tofile, os.stat(tofile).st_size))

    if len(filelist) > 0:
        archivefiles(filelist)
        LOG.info(f"NGAS file list: {filelist}")
    else:
        logger.warning("No SDM tables or BDF files to process.")


def generate_science_product_locator(telescope, filetype):
    """
    Generate unique science product locator for science products
    :param telescope: the telescope
    :param filetype: the type of the science product
    :return: science product locator
    """

    id = uuid.uuid4()
    science_product_locator = f"uid://{telescope.lower()}/{filetype}/{id}"
    return science_product_locator


def generate_uuids_for_ngas(telescope, filetype, extension):
    """
    Generate unique uuids for ngas storage
    :param filename: the current name of the file
    :return: new file name
    """

    id = uuid.uuid4()
    file = f"uid____{telescope}_{filetype}_{id}.{extension}"
    return file


def archivefiles(archivefilelist):
    filecount = 0
    for archivefile in archivefilelist:
        # archivefile 0:filename 1: file size on disk
        file_name = archivefile[0]
        expected_file_size = archivefile[1]
        # And away we go:
        starttime = time.time()
        NGASHOST = random_ngas_host()
        filecount += 1
        LOG.info(f"Ingesting: {file_name}")
        ngascmd = NGASHOST + archivecommand["archive"] + "?filename=file://" + quote_plus(file_name)
        LOG.info(ngascmd)
        response = requests.get(ngascmd)
        returncode = response.status_code
        archivemsg = "Archive request ({}): {} [{}] in {}s".format(
            filecount, ngascmd, returncode, time.time() - starttime
        )  # fix units
        LOG.info(archivemsg)
        doc = jxmlease.parse(response.text)
        if returncode == 200:
            ngas_file_size = int(doc["NgamsStatus"]["DiskStatus"]["FileStatus"].get_xml_attr("FileSize"))
            LOG.debug(f"Ingested file size: {ngas_file_size} Expected file size: {expected_file_size}")
            if ngas_file_size < expected_file_size:
                LOG.error(
                    f"NGAS ingested file size {ngas_file_size} indicates incomplete ingestion (expected file size {expected_file_size})"
                )
                exit(-1)
            elif ngas_file_size > expected_file_size:
                LOG.info(
                    f"NGAS ingested file size {ngas_file_size} greater than expected {expected_file_size}.  Weird, but continuing."
                )
            # If they're equal, we're quite happy.
        else:
            error_message = doc["NgamsStatus"]["Status"].get_xml_attr("Message")
            LOG.error(f"NGAS ingestion of {archivefile[0]} failed. Return code: {returncode} Message: {error_message}")
            exit(-1)


def sha256(fname):
    hash_sha256 = hashlib.sha256()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_sha256.update(chunk)
    return hash_sha256.hexdigest()


def test_uuid(telescope, filetype, extension):
    print(generate_uuids_for_ngas(telescope, filetype, extension))


if __name__ == "__main__":
    generate_science_product_locator("EVLA", "execblock")
